#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include "fifo.h"


/*******************/
/* Concurrent FIFO */
/*******************/


fifo_t *alloc_fifo(int sz) {
  fifo_t *r = malloc_check(sizeof(*r)) ;
  r->fst = r->lst = r->nitems = 0 ;
  r->sz = sz ;
  r->t = calloc(sz,sizeof(*r->t)) ;
  if (!r->t) {
    perror("calloc") ;
    exit(2) ;
  }
  r->fst_mutex = alloc_mutex() ;
  r->lst_mutex = alloc_mutex() ;
  r->is_empty = alloc_cond() ;
  r->is_full = alloc_cond() ;
  return r ;
}

void free_fifo(fifo_t *f) {
  free_cond(f->is_empty) ;
  free_cond(f->is_full) ;
  free_mutex(f->fst_mutex) ;
  free_mutex(f->lst_mutex) ;
  free(f->t) ;
  free(f) ;
}

/* real operations */

#define POSTINCR(p) __sync_fetch_and_add(p,1)

void put(fifo_t *f,void *z) {
  lock_mutex(f->lst_mutex) ;
  while (GET(&f->nitems) == f->sz) {
    wait_cond(f->is_full,f->lst_mutex) ;
  }
  f->t[f->lst] = z ;
  f->lst++ ; f->lst %= f->sz ;
  int was_empty = POSTINCR(&f->nitems) == 0 ;
  unlock_mutex(f->lst_mutex) ;  
  if (was_empty) {
    lock_mutex(f->fst_mutex) ;
    broadcast_cond(f->is_empty) ;
    unlock_mutex(f->fst_mutex) ;
  }
}

#define POSTDECR(p) __sync_fetch_and_add(p,-1)

void *get(fifo_t *f) {
  void *r ;
  lock_mutex(f->fst_mutex) ;
  while (GET(&f->nitems) == 0) {
    wait_cond(f->is_empty,f->fst_mutex) ;
  }
  r = f->t[f->fst] ;
  f->fst++ ; f->fst %= f->sz ;
  int was_full = POSTDECR(&f->nitems) == f->sz ;
  unlock_mutex(f->fst_mutex) ;
  if (was_full) {
    lock_mutex(f->lst_mutex) ;
    broadcast_cond(f->is_full) ;
    unlock_mutex(f->lst_mutex) ;
  }
  return r ;
}
