#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include "fifo.h"

static int verbose = 0 ;


/* Barrier */

#ifdef MY
#include"sense_barrier.c"
#else
typedef pthread_barrier_t barrier_t ;

inline static barrier_t *alloc_my_barrier(unsigned int sz) {
  return alloc_barrier(sz) ;
}

inline static void free_my_barrier(barrier_t *p) {
  free_barrier(p) ;
}

inline static void wait_my_barrier(barrier_t *p,int *sense) {
  (void)wait_barrier(p) ;
}
#endif



typedef struct {
  int id ;
  int size ;
  barrier_t *b ;
  volatile int *x,*y;
} ctx_t ;

void *P1(void *_p) {
  ctx_t *p = _p ;
  int sense = 0 ;
  int size = p->size ;
  for (int k = 0 ; k < size ; k++) {
    wait_my_barrier(p->b,&sense) ;
    p->x[k] = 1 ;
    p->y[k] = 1 ;
    //    wait_my_barrier(p->b,&sense) ;    
  }
  return NULL ;
}

void *P2(void *_p) {
  ctx_t *p = _p ;
  int sense = 0 ;
  int size = p->size ;
  int *t = calloc_check(size, sizeof(t[0]));
  for (int k = 0 ; k < size ; k++) {
    wait_my_barrier(p->b,&sense) ;
    p->y[k] = 2 ;
    int r = p->x[k] ;
    //    wait_my_barrier(p->b,&sense) ;    
    t[k] = r ;
  }
  return t ;
}

void run(int size) {
  barrier_t *b = alloc_my_barrier(2) ;

  for ( ; ; ) {
    int *x = calloc(size,sizeof(x[0])) ;
    int *y = calloc(size,sizeof(y[0])) ;

    pthread_t th1,th2 ;
    ctx_t a1,a2 ;
    
    a1.id = 1 ; a1.size = size ; a1.b = b ;  a1.x = x ; a1.y = y ;
    create_thread(&th1,P1,&a1) ;
    a2.id = 2 ; a2.size = size ; a2.b = b ;  a2.x = x ; a2.y = y ;
    create_thread(&th2,P2,&a2) ;

    (void)join_thread(&th1) ;
    int *r2 = join_thread(&th2) ;
    int count[3][2] ;

    for (int k1 = 0 ; k1 < 3 ; k1++)
      for (int k2 = 0 ; k2 < 2 ; k2++)
        count[k1][k2] = 0 ;

    for (int k = 0 ; k < size ; k++) {
      count[y[k]][r2[k]]++ ;
    }
    printf("**********\n") ;
    int nz = 0 ;
    for (int k1 = 1 ; k1 < 3 ; k1++)
      for (int k2 = 0 ; k2 < 2 ; k2++) {
        int c = count[k1][k2] ;
        if (c > 0) nz++ ;
        printf("%-5i> y=%i r2=%i\n",c,k1,k2) ;
      }
    free(x) ; free(y) ; free(r2) ;
    if (nz == 4) return ;
  }
  free_my_barrier(b) ;

}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 1024 ;

  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    size = atoi(argv[0]) ;
  }
  run(size) ;
  return 0 ;
}
