#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include "basic.h"
#include "stack.h"


/* Nodes */

static stack_node_t *alloc_stack_node(void *val) {
  stack_node_t *r = malloc_check(sizeof(*r)) ;
  r->val = val ; r->next = NULL ;
  return r ;
}

static void free_stack_node(stack_node_t *p) { free(p) ; }

stack_t *alloc_stack(int nprocs) {
  stack_t *r =  malloc_check(sizeof(*r)) ;
  r->top = NULL ;
  r->mutex = alloc_mutex() ;
  return r ;
}

void free_stack(stack_t *p) {
  free_mutex(p->mutex) ;
  free(p) ;
}

void push(stack_t *p, void *v) {  
  stack_node_t *q = alloc_stack_node(v) ;
  lock_mutex(p->mutex) ;
  stack_node_t *old = p->top ;
  q->next = old ;
  p->top = q ;
  unlock_mutex(p->mutex) ;
}

void *pop(stack_t *p,int nprocs) {
  lock_mutex(p->mutex) ;  
  stack_node_t *q = p->top ;
  if (q == NULL) {
    unlock_mutex(p->mutex) ;
    return NULL ;
  } else {
    p->top = q->next ;
    unlock_mutex(p->mutex) ;
    void *r = q->val ;
    free_stack_node(q) ;
    return r ;
  }
}

