#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include "basic.h"

enum  action_t { UNLOCK, LOCK, FETCH, };

static enum action_t action = UNLOCK ;
static int n = 10000000;
static volatile int sum = 0 ;
static volatile int start = 0 ;

static pthread_mutex_t *mutex ;

void *f_locked(void *p) {
  while (!start) ;
  for (int k = 0 ; k < n ; k++) {
    lock_mutex(mutex) ;
    sum++ ;
    unlock_mutex(mutex) ;
  }
  return NULL ;
}

void *f_fetch_and_add(void *p) {
  while (!start) ;
  for (int k = 0 ; k < n ; k++) {
    (void)__sync_fetch_and_add(&sum,1) ;
  }
  return NULL ;
}

void *f_unlocked(void *p) {
  while (!start) ;
  for (int k = 0 ; k < n ; k++) {
    sum++ ;
  }
  return NULL ;
}

void run(int nprocs) {
  mutex = alloc_mutex() ;
  sum = 0 ;
  start = 0 ;

  void *(*f)(void *) = f_unlocked ;
  if (action == LOCK) f = f_locked ;
  else if (action == FETCH) f = f_fetch_and_add ;
  
  pthread_t th[nprocs];
  for (int k = 0 ; k < nprocs ; k++) create_thread(&th[k],f,NULL) ;
  start = 1 ;
  for (int k = 0 ; k < nprocs ; k++) (void)join_thread(&th[k]) ;
  printf("%i\n",sum) ;
}


void usage(char *pgm)
{
  fprintf(stderr,"usage: %s [-jN] [n]\n",pgm);
  exit(-1);
}

int main (int argc, char **argv) {
  char *pgm = argv[0] ;
  char *arg = 0 ;
  int nprocs = 2 ;
  while (--argc) {
    arg = *++argv;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'u') action = UNLOCK ;
      else if (*arg == 'l') action = LOCK ;
      else if (*arg == 'f') action = FETCH ;
      else if (*arg == 'p' || *arg == 'j') {
        if (arg[1]) {
          nprocs = atoi(arg+1) ;
        } else {
          usage(pgm) ;
        }
      } else usage(pgm) ;
    } else break;
  }
  if (argc >= 1) n = atoi(arg) ;
  run(nprocs) ;
  return 0 ;
}
