#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include "fifo.h"

static int verbose = 0 ;



/* Barrier */

typedef struct {
  unsigned int nprocs ;
  volatile unsigned int count ;
} barrier_t ;


barrier_t *alloc_my_barrier(unsigned int nprocs) {
  barrier_t *p = malloc_check(sizeof(*p)) ;
  p->nprocs = p->count = nprocs ;
  return p ;
}

void free_my_barrier(barrier_t *p) {
  free(p) ;
}

inline static void show(barrier_t *b, int lim, int id) {
  if (verbose) { 
    fprintf(stdout,"%02i [%i-%i]\n",id,b->count,lim) ;
    fflush(stdout) ;
  }
}
void wait_my_barrier(barrier_t *p,int id) {
  (void)DECR(&p->count) ;
  while (p->count != 0) show(p,0,id);
}

typedef struct {
  int m ;
  barrier_t *b ;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;

void *runner(void *_p) {
  ctx_t *p = _p ; common_t *q = p->common ;
  for (int k = 0 ; k < q->m ; k++) {
    wait_my_barrier(q->b,p->id) ;
    printf("<%i>",p->id) ; fflush(stdout) ;
    (void)wait_my_barrier(q->b,p->id) ;
    if (p->id == 0)  printf("\n") ;
  }
  return NULL ;
}

void run(int n, int m) {
  common_t c ;
  c.b = alloc_my_barrier(n) ;
  c.m = m ;
  pthread_t th[n] ; ctx_t a[n] ;
  for (int k = 0 ; k < n ; k++) {
    ctx_t *p = &a[k] ;
    p->id = k ; p->common = &c ;
    create_thread(&th[k],runner,p) ;
  }

  for (int k = 0 ; k < n ; k++) {
    (void)join_thread(&th[k]) ;
  }
  free_my_barrier(c.b) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n] [m]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int n = 2 ;
  int m = 10 ;

  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    n = atoi(argv[0]) ;
  } else if (argc > 1) {
    n = atoi(argv[0]) ;
    m = atoi(argv[1]) ;
  }  
  run(n,m) ;
  return 0 ;
}
