#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include "fifo.h"

static int verbose = 0 ;
static int n = 2 ;
static int m = 1 ;



/* Barrier */

typedef struct {
  unsigned int nprocs ;
  volatile unsigned int count ;
  volatile int go ;
} barrier_t ;

#define DECR(x) __sync_add_and_fetch(x,-1)
#define INCR(x) __sync_add_and_fetch(x,1)
#if 0
#define GET(x)  __sync_add_and_fetch(x,0)
#define INVERSE(x) __sync_xor_and_fetch(x,1)
#else
#define GET(p)  *(p)
#define INVERSE(p)  *(p) = !*(p)
#endif

barrier_t *alloc_my_barrier(unsigned int nprocs) {
  barrier_t *p = malloc_check(sizeof(*p)) ;
  p->nprocs = p->count = nprocs ;
  p->go = 0 ;
  return p ;
}

void free_my_barrier(barrier_t *p) {
  free(p) ;
}

void wait_my_barrier(barrier_t *p) {
  int rem = DECR(&p->count) ;
  if (rem == 0) {
    p->go = 1 ;
    (void)INCR(&p->count) ;
    while (GET(&p->count) != p->nprocs) ;
    p->go = 0 ;
  } else {
    while (!GET(&p->go)) ;
    (void)INCR(&p->count) ;
    while (GET(&p->go)) ;
  }
}

static barrier_t *b1,*b2 ;

typedef struct {
  int id ;
} ctx_t ;

void *runner(void *_p) {
  ctx_t *p = _p ;
  for (int k = 0 ; k < m ; k++) {
    wait_my_barrier(b1) ;
    printf("<%i>",p->id) ; fflush(stdout) ;
    (void)wait_my_barrier(b1) ;
    if (p->id == 0)  printf("\n") ;
  }
  return NULL ;
}

void run(void) {
  b1 = alloc_my_barrier(n) ;
  b2 = alloc_my_barrier(n) ;
  pthread_t th[n] ;
  ctx_t arg[n] ;
  for (int k = 0 ; k < n ; k++) {
    arg[k].id = k ;
    create_thread(&th[k],runner,&arg[k]) ;
  }

  for (int k = 0 ; k < n ; k++) {
    (void)join_thread(&th[k]) ;
  }
  free_my_barrier(b2) ;
  free_my_barrier(b1) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n] [m]\n",pgm) ;
}
int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  fprintf(stderr,"argc=%i\n",argc) ;
  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    n = atoi(argv[0]) ;
    fprintf(stderr,"SET '%s' n: %i\n",argv[0],n) ;
  } else if (argc > 1) {
    n = atoi(argv[0]) ;
    m = atoi(argv[1]) ;
  }  
  if (verbose) fprintf(stderr,"n=%i, m=%i\n",n,m) ;
  run() ;
  return 0 ;
}
